import logging
logger = logging.getLogger(__name__)
logger.debug("Loaded " + __name__)

import json
from bson.objectid import ObjectId
from datetime import datetime

from kafka import KafkaProducer
from kafka import KafkaConsumer
from kafka import TopicPartition


def dict_to_binary(the_dict):
	binary = ' '.join(format(ord(letter), 'b') for letter in the_dict)
	return binary

def binary_to_dict(the_binary):
	jsn = ''.join(chr(int(x, 2)) for x in the_binary.split())
	return jsn

def kafka_to_dict(kafka_msg):
	msg = json.loads(binary_to_dict(kafka_msg.value))
	kafka_msg_id = "{id}:{topic}:{partition}:{offset}".format(**{ "id":msg["_id"],"offset":kafka_msg.offset, "partition": kafka_msg.partition, "topic":kafka_msg.topic })
	msg["_kafka__id"]= kafka_msg_id
	return msg
	
def dict_to_kafka(output,source_data):
	for data in source_data:
		if output["source_id"] == data["_id"]:
			output["_kafka_source_id"] = data["_kafka__id"]
			break
	kafka_msg = dict_to_binary(json.dumps(output))
	return kafka_msg

# TODO: Move/Add formatOutput to behaviour base class 
# Created following fields in output dict if missing:
# _id,_created,_updated,source_id,_type,_producer
def formatOutput(output,behavior,source_data): 
	if "_id" not in output: output["_id"] = str(ObjectId())
	if "_updated" not in output: output["_updated"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
	if "_type" not in output: output["_type"] = "BEHAVIOUR"		#TODO take from behavior object
	if "_producer" not in output: output["_producer"] = "{}:{}:{}".format(behavior.__class__.__name__,"1.0",behavior.id) #name:version:id #TODO take version from behaviour

	# Source chaining for stream
	if "source_id" not in output: 
		if source_data: # Select rightmost consumer
			output["source_id"] = source_data[-1]["_id"]
		else:  # This is Producer
			output["source_id"] = output["_id"]
	if "_created" not in output: 
		if output["source_id"] is None or output["source_id"] == output["_id"]:
			output["_created"] = output["_updated"]
		else:
			for data in source_data:
				if output["source_id"] == data["_id"]:
					output["_created"] = data["_created"]
					break

	if "_created" not in output: 		
		logger.info("{} | source_id  {} not found for id {}".format(output["_producer"],output["source_id"],output["_id"]))
		output["_created"] = output["_updated"]
		
	return output

class KafkaConnector(object):
	Type = "KafkaConnector"
	def __init__(self, Behaviour, producer_topic=None, consumer_topic=None, consumer_topic2=None, kafka_broker="localhost:9092", sync_consumer=True, auto_offset_reset='largest'):
	# def __init__(self, Behaviour):
		super(KafkaConnector, self).__init__()

		self.behavior = Behaviour
		self.producer_topic = producer_topic
		self.consumer_topic = consumer_topic
		self.consumer_topic2 = consumer_topic2
		self.sync_consumer = sync_consumer
		self.kafka_api_version = (2, 12, 2)

		logger.info("=" * 20)
		logger.info("Kafka INIT Config : ")
		logger.info("Behaviour : " + str(Behaviour))
		logger.info("producer_topic : " + str(producer_topic))
		logger.info("consumer_topic : " + str(consumer_topic))
		logger.info("consumer_topic2 : " + str(consumer_topic2))
		logger.info("sync_consumer : " + str(sync_consumer))
		logger.info("auto_offset_reset : " + str(auto_offset_reset))
		logger.info("=" * 20)

		if(producer_topic):
			self.producer = KafkaProducer(bootstrap_servers=kafka_broker, max_request_size=20000000)
		else:
			self.producer = None
		
		if(consumer_topic):
			self.consumer = KafkaConsumer(consumer_topic, bootstrap_servers=kafka_broker, auto_offset_reset=auto_offset_reset)
			self.consumer.poll()
		else:
			self.consumer = None

		if(consumer_topic2):
			self.consumer2 = KafkaConsumer(consumer_topic2, bootstrap_servers=kafka_broker, auto_offset_reset=auto_offset_reset)
			self.consumer2.poll()
		else:
			self.consumer2 = None

		


	def run(self):
		while True:
			if(self.consumer): # Check at least primary consumer is present
				logger.info("Consumed | {} | Topic : {}".format(self.behavior.__class__.__name__, self.consumer_topic))
				kafka_msg = next(self.consumer)
				msg = kafka_to_dict(kafka_msg)
			else:
				msg = None

			if(self.consumer2): # check for two consumers		
				try:
					if(self.sync_consumer):
						kafka_msg = next(self.consumer2)
						msg2 = kafka_to_dict(kafka_msg)
						assert msg2["_id"] == msg["source_id"]
					else:
						msg2_raw = self.consumer2.poll(max_records=1)

						if msg2_raw:
							msg2 = kafka_to_dict(msg2_raw.values()[0][0])							
						else:
							msg2 = None
				except AssertionError:

					logger.info("Syncing Partition...")
					kafka_source_id = msg["_kafka_source_id"]			#"{id}:{topic}:{partition}:{offset}"
					topicName = kafka_source_id.split(":")[-3] 			# 3rd last 
					partitionName = int(kafka_source_id.split(":")[-2]) # 3rd last
					offset =  int(kafka_source_id.split(":")[-1])
					partition = TopicPartition(topic=topicName, partition=partitionName) 

					logger.debug("Partition : " + str(partition))

					self.consumer2.seek(partition,offset)
					msg2 = kafka_to_dict(next(self.consumer2))

				output = self.behavior.run(msg, msg2)
			elif(self.consumer): # One consumer only
				output = self.behavior.run(msg)
			else: # Not even primary consumer present, producer only behaviour
				output = self.behavior.run()
			
			# Transform output to fill missing fields
			if output:
				source_data = []
				if self.consumer: source_data.append(msg)
				if self.consumer2: source_data.append(msg2)
				output=formatOutput(output,self.behavior,source_data)

			if(self.producer):
				logger.info("Produced | {} | Topic : {}".format(self.behavior.__class__.__name__, self.producer_topic))
				if(output):
					self.producer.send(topic=self.producer_topic, value=dict_to_kafka(output,source_data))



import logging
logger = logging.getLogger(__name__)
logger.debug("Loaded " + __name__)

import datetime
from flask import current_app as app

from influxdb import InfluxDBClient
clients = {}

def init_influxdb_client(db):
    if db not in clients.keys():
        clients[db] = InfluxDBClient(
            app.config["INFLUX_HOST"], 
            app.config["INFLUX_PORT"],
            app.config["INFLUX_USER"],
            app.config["INFLUX_PASSWORD"],
            db
        )
    return clients

def get_current_time():
    # return datetime.datetime.utcnow().isoformat()
    return datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")

def write_points(db, measurement, elements):
    init_influxdb_client(db)
    for element in elements:
        if "time" not in element:
            element["time"] = get_current_time()
        element["measurement"] = measurement
    logger.debug("Write points: {0}".format(elements))
    clients[db].write_points(elements)



import logging
logger = logging.getLogger(__name__)
logger.debug("Loaded " + __name__)

class Behaviour(object):
    def __init__(self, Config):
        self.id = Config["_id"]
        self.enabled = True
        self.Config  = Config
    
    def run(self, data, params):
        return(data)



import logging
logger = logging.getLogger(__name__)
logger.debug("Loaded " + __name__)

import os
import operator

from eazyserver.core.utils import list_routes, get_app_config_as_dict, toJSON

from flask import Blueprint, jsonify, Response, make_response, request, \
    current_app as app

index_bp = Blueprint('eazy_index_route', __name__)

@index_bp.route('/', methods=['GET'])
def index():
    response = {
        "_links": {
            "all": list_routes()
        },
        "_meta": {
            "build": {
                "href": os.environ.get('CI_BUILD_LINK', '-')
            }
        }
    }
    return jsonify(response)



# -*- coding: utf-8 -*-

"""Console script for eazyserver."""
import sys
import click

@click.group()
def cli():
    """Console script for eazyserver."""
    click.echo("Replace this message by putting your code into "
               "eazyserver.cli.cli")
    click.echo("See click documentation at http://click.pocoo.org/")

@cli.command()
def run():
    click.echo('Run subcommand')

@cli.command()
def start():
    click.echo('Start subcommand')


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



import logging
logger = logging.getLogger(__name__)
logger.debug("Loaded " + __name__)

# from ..core.function_mapper import FunctionMapper 
# mongo_pipelines = FunctionMapper()

from .function_mapper import *
from .utils import *
from .influxdb_connector import *
from .kafka_connector import KafkaConnector
from .behaviour_base import Behaviour


# -*- coding: utf-8 -*-

"""Top-level package for eazyserver."""

import os
BASE_DIR=os.path.dirname(os.path.abspath(__file__))
SETTINGS_INI = os.path.join(BASE_DIR, 'settings.ini')
LOGGER_CONFIG = os.path.join(BASE_DIR, 'logger.conf')
# VERSION_FILE = os.path.join(BASE_DIR, 'VERSION')
# with open('VERSION') as version_file:
#     version = version_file.read().strip()


__author__ = """Ashutosh Mishra"""
__email__ = 'ashutoshdtu@gmail.com'
__version__ = '0.2.7'


import logging
import logging.config
logging.config.fileConfig(LOGGER_CONFIG)
logger = logging.getLogger(__name__)
logger.debug("Loaded " + __name__)

from .eazyserver import Eazy



import logging
logger = logging.getLogger(__name__)
logger.debug("Loaded " + __name__)

import os
import operator

from flask import Blueprint, jsonify, Response, make_response, request, \
    current_app as app

health_bp = Blueprint('eazy_health_routes', __name__)

@health_bp.route('/health', methods=['GET'])
def health():
    response = {
        "status": "UP",
        "checks": []
    }
    return jsonify(response)

@health_bp.route('/ready', methods=['GET'])
def ready():
    response = {
        "status": "UP",
        "checks": []
    }
    return jsonify(response)



import logging
logger = logging.getLogger(__name__)
logger.debug("Loaded " + __name__)

import json
from ..rpc import methods

@methods.add
def __list_methods():
    result = []
    for method in methods:
        result.append(str(method))
    return result

@methods.add
def __ping():
    logger.debug("Inside Ping Pong")
    return 'pong'



import logging
logger = logging.getLogger(__name__)
logger.debug("Loaded " + __name__)

import json
import sys
if sys.version_info[0] >= 3:
    from html.parser import HTMLParser
else:
    import HTMLParser
from bson import json_util
from bson.son import SON
from bson.objectid import ObjectId
from eve.utils import str_to_date, date_to_rfc1123
from flask import current_app as app

def bson_to_json(_object):
    return json.loads(json.dumps(_object, default=json_util.default))

def json_to_bson(_object):
    return json_util.loads(json.dumps(_object))

def toJSON(_object):
    return json.dumps(_object, default= lambda o: o.__str__(), 
        sort_keys=True)

def unescape(s):
    h= HTMLParser.HTMLParser()
    return h.unescape(s)

def get_app_config_as_dict():
    response = {}
    for key in app.config.iterkeys():
        response[key] = app.config[key]
    return response

def list_routes():
    routes = []
    for rule in app.url_map.iter_rules():
        route = {
            "title": rule.endpoint,
            "href": unescape( str(rule) ),
            "methods": sorted(rule.methods)
        }
        routes.append(route)
    return routes
